// Copyright (C) 2004  EPITA Research and Development Laboratory
//
// This file is part of the Olena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
// Boston, MA 02110-1301, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.
#ifndef OLN_UTILS_QSORT_HH
# define OLN_UTILS_QSORT_HH
# include <vector>
# include <oln/core/abstract/image.hh>
# include <oln/core/abstract/image_with_type_with_dim.hh>

# include <algorithm>
# include <iostream>

namespace oln
{
  namespace utils
  {
    namespace internal
    {
      /*!
      ** \brief Functor to sort compare 2 points using an image.
      **
      ** \param I Exact type of the image.
      */
      template <typename I>
      struct cmp_pt_value
      {
	typedef oln::abstract::non_vectorial_image<I>	im_type; ///< Type of the image.
	typedef oln_point_type(I)			point_type; ///< Type of the considered points.

	/*!
	** \brief Constructor.
	**
	** \arg im Image to use for comparison.
	*/
	cmp_pt_value(const im_type &im): im_(im)
	{
	}

	/*!
	** \brief Main method.
	**
	** \arg i The first point.
	** \arg j The second point.
	** \return im[i] < im[j]
	** \warning In  the case where im_[j] =  im_[i] the comparison
	** is  done on  the point  type components  (for compatibility
	** purpose with distributed sort).
	*/
	bool operator()(const point_type &i, const point_type &j)
	{
	  if (im_[i] == im_[j])
	    for (unsigned t(0); t < point_type::dim; ++t)
	      if (i.nth(t) != j.nth(t))
		return i.nth(t) < j.nth(t);
	  return im_[j] > im_[i];
	}

      protected:
	const im_type	&im_; ///< Image to use for comparison.
      };

      /*!
      ** \brief Functor to sort compare 2 points using an image
      **
      ** \param I Exact type of the image.
      */
      template <typename I>
      struct cmp_pt_value_inv
      {
	typedef oln::abstract::non_vectorial_image<I>	im_type; ///< Type of the image.
	typedef oln_point_type(I)			point_type; ///< Type of the considered points.

	/*!
	** \brief Constructor.
	**
	** \arg im Image to use for comparison.
	*/
	cmp_pt_value_inv(const im_type &im): im_(im)
	{
	}

	/*!
	** \brief Main method.
	**
	** \arg i The first point.
	** \arg j The second point.
	** \return im[i] > im[j]
	** \warning In  the case where im_[j] =  im_[i] the comparison
	** is  done on  the point  type components  (for compatibility
	** purpose with distributed sort).
	*/
	bool operator()(const point_type &i, const point_type &j)
	{
	  if (im_[i] == im_[j])
	    for (unsigned t(0); t < point_type::dim; ++t)
	      if (i.nth(t) != j.nth(t))
		return i.nth(t) < j.nth(t);
	  return im_[i] > im_[j];
	}

      protected:
	const im_type	&im_; ///< Image to use for comparison.
      };
    } // !internal

    /*!
    ** \brief Sort points of an image in increasing order.
    **
    ** \param I Exact type of the image.
    **
    ** \arg im The input image.
    ** \arg v The result.
    */
    template<class I>
    void
    qsort(const oln::abstract::non_vectorial_image<I>& im,
	  std::vector<oln_point_type(I)> &v)
    {
      oln_iter_type(I)	it(im);

      for_all(it)
	v.push_back(oln_point_type(I)(it));
      internal::cmp_pt_value<I> c(im);

      std::sort(v.begin(), v.end(), c);
    }

    /*!
    ** \brief Sort points of an image in decreasing order.
    **
    ** \param I Exact type of the image.
    **
    ** \arg im The input image.
    ** \arg v The result.
    */
    template<class I>
    void
    qsort_inv(const oln::abstract::non_vectorial_image<I>& im,
	      std::vector<oln_point_type(I)> &v)
    {
      oln_iter_type(I)	it(im);

      for_all(it)
	v.push_back(oln_point_type(I)(it));

      internal::cmp_pt_value_inv<I> c(im);

      std::sort(v.begin(), v.end(), c);
    }

    /*! Select statically the good qsort.
    **
    ** \param reverse If the sort should be reverted or not.
    */
    template <bool reverse>
    struct select_q_sort
    {
      template <class I>
      void
      operator ()(const oln::abstract::non_vectorial_image<I>& im,
		  std::vector<oln_point_type(I)> &v)
      {
	qsort_inv(im, v);
      }
    };

    /*! Select statically the good qsort.
    **
    ** Specialisation.
    */
    template <>
    struct select_q_sort<true>
    {
      template <class I>
      void
      operator ()(const oln::abstract::non_vectorial_image<I>& im,
		  std::vector<oln_point_type(I)> &v)
      {
	qsort(im, v);
      }
    };
  } // !utils
} // !oln

#endif // !QSORT
